/***************************************************************************
*   Copyright (C) 2010 by
*    Kai Lindholm <megantti@gmail.com>
*   
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
* 
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*   
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the
*   Free Software Foundation, Inc.,
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
****************************************************************************/

#include "particle.h"
#include "settings.h"


using sf::operator+;
using sf::operator-;
using sf::operator*;


particle::particle(double _life, sf::Color _col, sf::Vector2f _pos, sf::Vector2f _vel, 
		sf::Vector2f _accel) :
	life(_life), col(_col), pos(_pos), vel(_vel), accel(_accel)
{
	SetPosition(pos);
}

void particle::update()
{
	life -= DT;
	col.a = life * 10.0f;
	move();

	SetPosition(pos);
}

void particle::move()
{
	pos += DT * vel;
	vel += DT * accel;
}

void particle::Render(sf::RenderTarget &target) const
{
	sf::Vector2f e ;
	e += vel * 0.4f;
	
	sf::Shape l = sf::Shape::Line(sf::Vector2f(), e, 2.0f, col);
	target.Draw(l);
}

particle_emitter::particle_emitter(particle_engine &e, double _rate) :
	particles(0), rate(_rate), maximum_age(0), age(0), engine(e)
{

}

void particle_emitter::update()
{
	particles += rate * DT;
	age += DT;
	for(int a = 0; a < particles; a++) {
		engine.add(create());
		particles -= 1.0f;
	}
}

float particle_emitter::uniform_dist(float cent, float r)
{
	return cent + 2 * r * (float(rand()) - float(RAND_MAX) / 2) / float(RAND_MAX);
}

particle *particle_emitter::create()
{
	return new particle(15.0f, sf::Color::Blue, sf::Vector2f());
}

particle_engine::~particle_engine()
{
	for(p_iter i = particles.begin(); i != particles.end(); i++) {
		delete *i;
	}
}


void particle_engine::update()
{
	for(p_iter i = particles.begin(); i != particles.end(); i++) {
		(*i)->update();
		if((*i)->dead()) {
			i = particles.erase(i);
		}
	}
}

void particle_engine::draw(sf::RenderTarget &target)
{
	for(p_iter i = particles.begin(); i != particles.end(); i++) {
		target.Draw(**i);
	}
}

void particle_engine::add(particle *p) 
{
	particles.push_back(p);
}
