/***************************************************************************
*   Copyright (C) 2010 by
*    Kai Lindholm <megantti@gmail.com>
*   
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
* 
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*   
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the
*   Free Software Foundation, Inc.,
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
****************************************************************************/

#ifndef _PARTICLE_H
#define _PARTICLE_H

#include <list>
#include <SFML/Graphics.hpp>

class particle : public sf::Drawable {
	public:
		particle(double life, sf::Color, sf::Vector2f pos, sf::Vector2f vel = sf::Vector2f(), 
				sf::Vector2f accel = sf::Vector2f());
		
		void update();
		bool dead() { return life <= 0 || pos.x >= 1300; }
		sf::Vector2f get_pos() { return pos; }
	protected:
		virtual void move();
		virtual void Render(sf::RenderTarget &) const;

		double life;
		sf::Color col;

		sf::Vector2f pos;
		sf::Vector2f vel;
		sf::Vector2f accel;
};

class particle_engine {
	public:
		~particle_engine();

		void add(particle*);

		void draw(sf::RenderTarget &);
		void update();
	private:
		std::list<particle*> particles;
		typedef std::list<particle*>::iterator p_iter;
};

class particle_emitter {
	public:
		particle_emitter(particle_engine &, double rate);
		void update();

		static float uniform_dist(float cent, float r);
		void set_maximum_age(double d) { maximum_age = d; }
		double get_age() { return age; }
		bool dead() { return age >= maximum_age; }
	protected:
		virtual particle* create();
	private:
		double particles;
		double rate;
		double maximum_age;
		double age;
		particle_engine &engine;
};


#endif

