/***************************************************************************
*   Copyright (C) 2010 by
*    Santtu Keskinen <laquendi@gmail.com>
*   
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
* 
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*   
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the
*   Free Software Foundation, Inc.,
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
****************************************************************************/

#include "world.h"
#include "spell.h"


spell_info_container::spell_info_container() {
	//cc_grow_mushroom
	infos.push_back(spell_info("grow_mushroom")
		.set_desc("Grow new mushroom. Every mushroom costs more energy than last one.")
		.set_target(spell_info::ground).set_range(100).set_hotkey('e')
		.set_e_cost(50)
	);
	//cc_connect
	infos.push_back(spell_info("connect_mushroom")
		.set_desc("Connect command center to a mushroom")
		.set_target(spell_info::mushroom).set_range(200).set_hotkey('c')
		.set_e_cost(30)
	);
	//cc_fertilize_str	
	infos.push_back(spell_info("fertilize_str")
		.set_desc("Boost attack mushrooms, harms tanks. Energy cost per selected mushroom")
		.set_hotkey('s').set_e_cost(30)
	);
	//cc_fertilize_mana
	infos.push_back(spell_info("fertilize_mana")
		.set_desc("Boost spore mushrooms, harms attacks. Energy cost per selected mushroom")
		.set_hotkey('m').set_e_cost(20)
	);
	//cc_fertilize_dur
	infos.push_back(spell_info("fertilize_dur")
		.set_desc("Boost tanks mushrooms, harms spores. Energy cost per selected mushroom")
		.set_hotkey('d').set_e_cost(30)
	);
	//mushroom_upg_to_attack
	infos.push_back(spell_info("upg_to_attack")
		.set_desc("Turns mushroom into attack mushroom")
		.set_hotkey('a').set_e_cost(30)
	);
	//mushroom_upg_to_spore
	infos.push_back(spell_info("upg_to_spore")
		.set_desc("Turns mushroom into spore mushroom")
		.set_hotkey('s').set_req_level(0)
	);
	//mushroom_upg_to_tank
	infos.push_back(spell_info("upg_to_tank")
		.set_desc("Turns mushroom into tank mushroom")
		.set_hotkey('t')
	);
	//mushroom_connect
	infos.push_back(spell_info("connect_mushroom")
		.set_desc("Connect two mushrooms")
		.set_target(spell_info::mushroom).set_hotkey('c').set_e_cost(35)
		.set_range(2)
	);
	//mushroom_multiply
	infos.push_back(spell_info("grow_mushroom")
		.set_desc("Grow new mushroom")
		.set_target(spell_info::ground).set_hotkey('e').set_range(1.2).set_e_cost(40)
	);
	//mushroom_give_energy
	infos.push_back(spell_info("mana_orb")
		.set_desc("Give energy to linked mushroom with 0.35 efficiency.")
		.set_e_cost(20).set_range(1337).set_g_cd(0)
		.set_hotkey('g').set_target(spell_info::mushroom_growing)
	);
	//mushroom_attack_wall
	infos.push_back(spell_info("attack_wall")
		.set_desc("Deal damage to wall")
		.set_target(spell_info::wall).set_hotkey('w').set_range(2)
	);
	//mushroom_target_ant
	infos.push_back(spell_info("target")
		.set_desc("Make mushroom attack selected ant")
		.set_target(spell_info::ant).set_hotkey('a').set_range(4)
	);
	//mushroom_paralyse
	infos.push_back(spell_info("paralyse")
		.set_desc("Paralyses target ant making it unable to move")
		.set_hotkey('p').set_target(spell_info::ant).set_range(1)
	);
	//mushroom_heal
	infos.push_back(spell_info("heal")
		.set_desc("Heal connected mushroom")
		.set_range(1337).set_e_cost(15)
		.set_target(spell_info::mushroom_growing).set_hotkey('h')
	);
	//mushroom_taunt
	infos.push_back(spell_info("taunt")
		.set_desc("Get nearby ants to target this mushroom")
		.set_hotkey('t').set_s_cd(40)
	);
	//unicorn_invisibility
	infos.push_back(spell_info("invisibility")
		.set_desc("Make mushrooms and command center invisible for some time")
		.set_range(150).set_s_cd(60)
		.set_target(spell_info::ground).set_hotkey('i')
	);
	//unicorn_tmp_dimension_door
	infos.push_back(spell_info("dimension_door")
		.set_desc("Teleport to a new location")
		.set_check_los(false).set_range(300).set_req_level(2)
		.set_target(spell_info::ground).set_hotkey('d')
	);
	//unicorn_mana_orb
	infos.push_back(spell_info("mana_orb")
		.set_desc("Give energy to mushrooms in area")
		.set_s_cd(40)
		.set_target(spell_info::ground).set_hotkey('o').set_req_level(2)
	);
	//unicorn_diagonal_spawn
	infos.push_back(spell_info("diagonal_spawn")
		.set_desc("Target spore mushroom. Spawns two mushrooms along all diagonals")
		.set_req_level(3).set_s_cd(90)
		.set_target(spell_info::mushroom).set_hotkey('a')
	);
	//unicorn_spawn_mushroom
	infos.push_back(spell_info("grow_mushroom")
		.set_desc("Spawn mushroom, has can't be casted too up")
		.set_req_level(4).set_y_limit(300).set_s_cd(15)
		.set_target(spell_info::ground).set_hotkey('e')
	);
	//unicorn_resurrect
	infos.push_back(spell_info("resurrect")
		.set_desc("Resurrect 10 dead mushrooms")
		.set_hotkey('r').set_req_level(4).set_s_cd(180)
	);
	//minotaur_attack_wall
	infos.push_back(spell_info("attack_wall")
		.set_desc("Deal damage to wall")
		.set_target(spell_info::wall).set_hotkey('w').set_range(70)
	);
	//griffin_eat_mushroom
	infos.push_back(spell_info("eat_mushroom")
		.set_desc("Eat mushroom for 500exp")
		.set_target(spell_info::mushroom).set_hotkey('e')
	);
}

void diagonal_spawn(sf::Vector2f dir, mushroom *spawner, world &w) {
	std::vector<mushroom*> &mushrooms = w.get_mushrooms();
	
	int max_tries = 4;
	int tries = 1;
	int success = 0;
	
	while (tries <= max_tries && success < 2) {
		sf::Vector2f pos = spawner->get_pos() + sf::Vector2f(dir.x*(40 * tries),
						dir.y*(40*tries));
		if(!w.out_of_bounds(pos) && w.get_map().los(spawner->get_center(), pos))
		{
			mushroom *m = new mushroom(w.anim_loader);
			m->set_pos(pos);
			
			if(!w.collision(*m)) {
				spawner->connect(m);
				m->connect(spawner);
				
				mushrooms.push_back(m);
				w.add_obj(m);
				success++;
			} else {
				delete m;
			}
		}
		tries++;
	}
}

bool spell::cast(selectable *caster, sf::Vector2f pos, world &w) const
{
	if(!can_cast(caster, pos, w)) 
		return false;
	
	sf::Vector2f caster_center = caster->get_center();
	mushroom *c_mushroom = dynamic_cast<mushroom*>(caster);
	creature *c_creature = dynamic_cast<creature*>(caster);
	command_center *c_cc = dynamic_cast<command_center*>(caster);
	
	mushroom *nearest_mushroom = w.nearest_mushroom(pos, SELECT_MUSHROOM_RANGE);
	mushroom *nearest_mushroom_growing = w.nearest_mushroom(pos, SELECT_MUSHROOM_RANGE, true);
//	command_center *nearest_cc = w.nearest_cc(pos, SELECT_CC_RANGE);
	ant *nearest_ant = w.nearest_ant(pos, SELECT_ANT_RANGE);

	// spell effects
	switch(t) {
		case cc_grow_mushroom: 
		{
			std::vector<mushroom*> &mushrooms = w.get_mushrooms();
			mushroom *m = new mushroom(w.anim_loader);
			m->set_pos(pos - sf::Vector2f(m->get_w()/2, m->get_h()/2));

			if(w.collision(*m)) {
				delete m;
				return false;
			}
			
			c_cc->connect_mushroom(m);
		
			mushrooms.push_back(m);
			w.add_obj(m);
		}
		break;
		case cc_connect:
		{
			if(!c_cc->is_connected(nearest_mushroom)) {
				c_cc->connect_mushroom(nearest_mushroom);

				goto success;
			}
			return false;
		}
		break;
		case cc_fertilize_str:
		{
			const std::list<mushroom*> &tree = c_cc->get_tree();
			if(tree.size()==0)
				return false;
			std::list<mushroom*>::const_iterator it;
			for(it = tree.begin(); it != tree.end(); it++) {
				if((*it)->get_type() == mushroom::m_attack)
					(*it)->change_modifier(7);
				if((*it)->get_type() == mushroom::m_tank)
					(*it)->change_modifier(-7);
			}
		}
		break;
		case cc_fertilize_mana:
		{
			const std::list<mushroom*> &tree = c_cc->get_tree();
			if(tree.size()==0)
				return false;
			std::list<mushroom*>::const_iterator it;
			for(it = tree.begin(); it != tree.end(); it++) {
				if((*it)->get_type() == mushroom::m_spore)
					(*it)->change_modifier(7);
				if((*it)->get_type() == mushroom::m_attack)
					(*it)->change_modifier(-7);
			}
		}
		break;
		case cc_fertilize_dur:
		{
			const std::list<mushroom*> &tree = c_cc->get_tree();
			if(tree.size()==0)
				return false;
			std::list<mushroom*>::const_iterator it;
			for(it = tree.begin(); it != tree.end(); it++) {
				if((*it)->get_type() == mushroom::m_tank)
					(*it)->change_modifier(7);
				if((*it)->get_type() == mushroom::m_spore)
					(*it)->change_modifier(-7);
			}
		}
		break;
		case mushroom_upg_to_attack: 
		{
			c_mushroom->grow(mushroom::m_attack);
		}
		break;
		case mushroom_upg_to_spore: 
		{
			c_mushroom->grow(mushroom::m_spore);
		}
		break;
		case mushroom_upg_to_tank: 
		{
			c_mushroom->grow(mushroom::m_tank);
		}
		break;
		case mushroom_connect:
		{
			if(!c_mushroom->is_connected(nearest_mushroom)) {
				nearest_mushroom->connect(c_mushroom);
				c_mushroom->connect(nearest_mushroom);

				goto success;
			}
			return false;
		}
		break;
		case mushroom_multiply:
		{
			std::vector<mushroom*> &mushrooms = w.get_mushrooms();
			mushroom *m = new mushroom(w.anim_loader);
			m->set_pos(pos - sf::Vector2f(m->get_w()/2, m->get_h()/2));

			if(w.collision(*m)) {
				delete m;
				return false;
			}

			c_mushroom->connect(m);
			m->connect(c_mushroom);
			mushrooms.push_back(m);
			w.add_obj(m);
		}
		break;
		case mushroom_give_energy:
		{
			std::list<mushroom*> l = c_mushroom->get_connections();
			std::list<mushroom*>::iterator it;
			for(it = l.begin(); it != l.end(); it++) {
				if((*it) == nearest_mushroom_growing) {
					nearest_mushroom_growing->add_energy(get_energy_cost(caster)*0.35);
					goto success;
				}
			}
			return false;
		}
		break;
		case mushroom_attack_wall:
		{
			tile & t = w.get_map().get_tile(sf::Vector2i(pos.x / TILE_W, pos.y / TILE_H));
			t.hit(w.anim_loader, 30);
		}
		break;
		case mushroom_target_ant:
		{
			c_mushroom->set_target(nearest_ant);
		}
		break;
		case mushroom_paralyse:
		{
			nearest_ant->paralyse(7.0f);
		}
		break;
		case mushroom_heal:
		{
			std::list<mushroom*> l = c_mushroom->get_connections();
			std::list<mushroom*>::iterator it;
			for(it = l.begin(); it != l.end(); it++) {
				if((*it) == nearest_mushroom_growing) {
					nearest_mushroom_growing->add_hp(10);
					goto success;
				}
			}
			return false;
		}
		break;
		case mushroom_taunt:
		{
			std::vector<ant*> &ants = w.get_ants();
			for(size_t a = 0; a < ants.size(); a++) {
				if(dist(c_mushroom->get_center(), ants[a]->get_center()) <= 200.0f) {
					ants[a]->set_target(c_mushroom);
				}
			}
		}
		break;
		case unicorn_invisibility:
		{
			std::list<selectable*> objs = w.get_m_and_cc_in_range(pos, 150);
			std::list<selectable*>::iterator it;
			for(it = objs.begin(); it != objs.end(); it++) {
				(*it)->set_invisible(45.0f);
			}
		}
		break;
		case unicorn_tmp_dimension_door:
		{
			sf::Vector2f cur_pos = caster->get_pos();
			w.obj_map_remove(caster);
			caster->set_pos(sf::Vector2f(pos.x-caster->get_w()/2, pos.y-caster->get_h()/2));
			if(w.collision(*caster)) {
				caster->set_pos(cur_pos);
				w.obj_map_add(caster);
				return false;
			}
			w.obj_map_add(caster);
			
			c_creature->set_move_pos(c_creature->get_center());
		}
		break;
		case unicorn_mana_orb:
		{	
			std::vector<mushroom*>::iterator it = w.get_mushrooms().begin();
			for(;it != w.get_mushrooms().end(); it++) {
				if(dist(pos, (*it)->get_center()) < 100)
					(*it)->add_energy(15.0f);
			}
		}
		break;
		case unicorn_diagonal_spawn:
		{
			if(nearest_mushroom->get_type() == mushroom::m_spore) {
				diagonal_spawn(sf::Vector2f(-1, -1), nearest_mushroom, w);
				diagonal_spawn(sf::Vector2f(-1, 1), nearest_mushroom, w);
				diagonal_spawn(sf::Vector2f(1, -1), nearest_mushroom, w);
				diagonal_spawn(sf::Vector2f(1, 1), nearest_mushroom, w);
			} else {
				return false;
			}
		}
		break;
		case unicorn_spawn_mushroom: 
		{
			std::vector<mushroom*> &mushrooms = w.get_mushrooms();
			mushroom *m = new mushroom(w.anim_loader);
			m->set_pos(pos - sf::Vector2f(m->get_w()/2, m->get_h()/2));
			
			if(w.collision(*m)) {
				delete m;
				return false;
			}
		
			mushrooms.push_back(m);
			w.add_obj(m);
		}
		break;
		case unicorn_resurrect:
		{
			int i = 0;
			while (i < 10 && w.has_deceased_m()) {
				std::pair<sf::Vector2f, mushroom::mushroom_type> dead = 
						w.get_deceased_m();
				mushroom *m = new mushroom(w.anim_loader);
				m->set_pos(dead.first);
				if(w.collision(*m)) {
					delete m;
				} else {
					i++;
					m->stop_upgrade();
					m->upgrade(dead.second);
					m->set_mode("stand");
					m->calculate_abilities();
					m->add_hp(m->get_max_hp());
					w.get_mushrooms().push_back(m);
					w.add_obj(m);
				}
			}
		}
		break;
		case minotaur_attack_wall:
		{
			tile & t = w.get_map().get_tile(sf::Vector2i(pos.x / TILE_W, pos.y / TILE_H));
			t.hit(w.anim_loader, 30);
		}
		break;
		case griffin_eat_mushroom:
		{
			w.disconnect_mushroom(nearest_mushroom);

			std::vector<mushroom*> &mushrooms = w.get_mushrooms();
			std::vector<mushroom*>::iterator it = mushrooms.begin();
			for(;it!=mushrooms.end(); it++)
			{
				if(nearest_mushroom==(*it))
				{
					w.delete_obj(*it);
					mushrooms.erase(it);
					break;
				}
			}
			w.get_player().gain_exp(w, 500);
		}
		break;
		default:
			return false;
	}

success:
	
	w.handler.cast_spell(caster, *this);

	return true;
}

